#import python modules
import numpy as np
import time

#import compiled fortran library
import ex3lib
import atomwrite

import sys
import os

#whether or not to use the 3d visualization
UseVisual = False
if UseVisual:
    #import custom visualization library
    import atomvis

# whether or not to run the production steps
Production = True

#set the random number seed; useful for debugging
#np.random.seed(342324)
np.random.seed()

#NOTE:
#everything below assumes unit atomic masses,
#such that forces = accelerations.


def LineSearch(Pos, Dir, dx, EFracTol, AtomType,
               Accel = 1.5, MaxInc = 10., MaxIter = 10000):
    """Performs a line search along direction Dir.
Input:
    Pos: starting positions, (N,3) array
    Dir: (N,3) array of gradient direction
    dx: initial step amount
    EFracTol: fractional energy tolerance
    AtomType: (N) array of atom type
    Accel: acceleration factor
    MaxInc: the maximum increase in energy for bracketing
    MaxIter: maximum number of iteration steps
Output:
    PEnergy: value of potential energy at minimum along Dir
    Pos: minimum energy (N,3) position array along Dir
"""
    #identify global variables
    global M, L, Cut
    
    #start the iteration counter
    Iter = 0

    #find the normalized direction    
    NormDir = np.clip(Dir, -1.e100, 1.e100)
    NormDir = NormDir / np.sqrt(np.sum(NormDir * NormDir))

    #take the first two steps and compute energies    
    Dists = [0., dx]
    PEs = [ex3lib.calcenergy(Pos + NormDir * x, M, L, Cut, AtomType) for x in Dists]
    
    #if the second point is not downhill in energy, back
    #off and take a shorter step until we find one
    while PEs[1] > PEs[0]:
        Iter += 1
        dx = dx * 0.5
        Dists[1] = dx
        PEs[1] = ex3lib.calcenergy(Pos + NormDir * dx, M, L, Cut, AtomType)
        
    #find a third point
    Dists = Dists + [2. * dx]
    PEs = PEs + [ex3lib.calcenergy(Pos + NormDir * 2. * dx, M, L, Cut, AtomType)]
    
    #keep stepping forward until the third point is higher
    #in energy; then we have bracketed a minimum
    while PEs[2] < PEs[1]:
        Iter += 1
            
        #find a fourth point and evaluate energy
        Dists = Dists + [Dists[-1] + dx]
        PEs = PEs + [ex3lib.calcenergy(Pos + NormDir * Dists[-1], M, L, Cut, AtomType)]

        #check if we increased too much in energy; if so, back off
        if (PEs[3] - PEs[0]) > MaxInc * (PEs[0] - PEs[2]):
            PEs = PEs[:3]
            Dists = Dists[:3]
            dx = dx * 0.5
        else:
            #shift all of the points over
            PEs = PEs[-3:]
            Dists = Dists[-3:]
            dx = dx * Accel
            
    #we've bracketed a minimum; now we want to find it to high
    #accuracy
    OldPE3 = 1.e300
    while True:
        Iter += 1
        if Iter > MaxIter:
            print "Warning: maximum number of iterations reached in line search."
            break
            
        #store distances for ease of code-reading
        d0, d1, d2 = Dists
        PE0, PE1, PE2 = PEs

        #use a parobolic approximation to estimate the location
        #of the minimum
        d10 = d0 - d1
        d12 = d2 - d1
        Num = d12*d12*(PE0-PE1) - d10*d10*(PE2-PE1)
        Dem = d12*(PE0-PE1) - d10*(PE2-PE1)
        if Dem == 0:
            #parabolic extrapolation won't work; set new dist = 0 
            d3 = 0
        else:
            #location of parabolic minimum
            d3 = d1 + 0.5 * Num / Dem
            
        #compute the new potential energy
        PE3 = ex3lib.calcenergy(Pos + NormDir * d3, M, L, Cut, AtomType)
        
        #sometimes the parabolic approximation can fail;
        #check if d3 is out of range < d0 or > d2 or the new energy is higher
        if d3 < d0 or d3 > d2 or PE3 > PE0 or PE3 > PE1 or PE3 > PE2:
            #instead, just compute the new distance by bisecting two
            #of the existing points along the line search
            if abs(d2 - d1) > abs(d0 - d1):
                d3 = 0.5 * (d2 + d1)
            else:
                d3 = 0.5 * (d0 + d1)
            PE3 = ex3lib.calcenergy(Pos + NormDir * d3, M, L, Cut, AtomType)
            
        #decide which three points to keep; we want to keep
        #the three that are closest to the minimum
        if d3 < d1:
            if PE3 < PE1:
                #get rid of point 2
                Dists, PEs = [d0, d3, d1], [PE0, PE3, PE1]
            else:
                #get rid of point 0
                Dists, PEs = [d3, d1, d2], [PE3, PE1, PE2]
        else:
            if PE3 < PE1:
                #get rid of point 0
                Dists, PEs = [d1, d3, d2], [PE1, PE3, PE2]
            else:
                #get rid of point 2
                Dists, PEs = [d0, d1, d3], [PE0, PE1, PE3]
                
        #check how much we've changed
        if abs(OldPE3 - PE3) < EFracTol * abs(PE3):
            #the fractional change is less than the tolerance,
            #so we are done and can exit the loop
            break
        OldPE3 = PE3

    #return the position array at the minimum (point 1)        
    PosMin = Pos + NormDir * Dists[1]
    PEMin = PEs[1]

    #if using visualization, update the display
    if UseVisual:
        if atomvis.Initialized:
            #update the positions
            atomvis.Update(PosMin)
        else:
            #initialize the visualization window
            atomvis.Init(PosMin)
        
    return PEMin, PosMin

        
def ConjugateGradient(Pos, dx, EFracTolLS, EFracTolCG, AtomType):
    """Performs a conjugate gradient search.
Input:
    Pos: starting positions, (N,3) array
    dx: initial step amount
    EFracTolLS: fractional energy tolerance for line search
    EFracTolCG: fractional energy tolerance for conjugate gradient
    AtomType: (N) array of atom type
Output:
    PEnergy: value of potential energy at minimum
    Pos: minimum energy (N,3) position array 
"""
    #identify global variables
    global M, L, Cut
    #initial search direction
    Forces = np.zeros_like(Pos)
    PE, Forces = ex3lib.calcenergyforces(Pos, M, L, Cut, Forces, AtomType)
    #print PE
    #print Forces
    Dir = Forces
    OldPE = 1.e300
    #iterative line searches
    while abs(PE - OldPE) > EFracTolCG * abs(PE):
        OldPE = PE
        PE, Pos = LineSearch(Pos, Dir, dx, EFracTolLS, AtomType)
        OldForces = Forces.copy()
        PE, Forces = ex3lib.calcenergyforces(Pos, M, L, Cut, Forces, AtomType)
        Gamma = np.sum((Forces - OldForces) * Forces) / np.sum(OldForces * OldForces)
        Dir = Forces + Gamma *  Dir
    return PE, Pos


def InitPositions(N, L):
    """Returns an array of initial positions of each atom,
placed on a cubic lattice for convenience.
Input:
    N: number of atoms
    L: box length
Output:
    Pos: (N,3) array of positions
"""
    #make the position array
    Pos = np.zeros((N,3), float)
    #compute integer grid # of locations for cubic lattice
    NLat = int(N**(1./3.) + 1.)
    LatSpac = L / NLat
    #make an array of lattice sites
    r = LatSpac * np.arange(NLat, dtype=float) - 0.5*L
    #loop through x, y, z positions in lattice until done
    #for every atom in the system
    i = 0
    for x in r:
        for y in r:
            for z in r:
                Pos[i] = np.array([x,y,z], float)
                #add a random offset to help initial minimization
                Offset = 0.1 * LatSpac * (np.random.rand(3) - 0.5)
                Pos[i] = Pos[i] + Offset
                i += 1
                #if done placing atoms, return
                if i >= N:
                    return Pos
    return Pos


def RescaleVelocities(Vel, T):
    """Rescales velocities in the system to the target temperature.
Input:
    Vel: (N,3) array of atomic velocities
    T: target temperature
Output:
    Vel: same as above
"""
    #recenter to zero net momentum (assuming all masses same)
    Vel = Vel - Vel.mean(axis=0)
    #find the total kinetic energy
    KE = 0.5 * np.sum(Vel * Vel)
    #find velocity scale factor from ratios of kinetic energy
    VScale = np.sqrt(1.5 * len(Vel) * T / KE)
    Vel = Vel * VScale
    return Vel  


def Berendsen(Vel, T, dt, tau):
    """Rescale velocities to target temperature using Berendsen thermostat.
    Input:
        Vel: (N,3) array of atomic velocities
        T: target temperature
        dt: MD time step
        tau: Berendsen thermostat coupling constant
    Output:
        Vel: same as above
"""
    #recenter to zero net momentum (assuming all masses same)
    Vel = Vel - Vel.mean(axis=0)
    #find the total kinetic energy
    KE = 0.5 * np.sum(Vel * Vel)
    #find the instantaneous temperature
    Tinst = 2*KE/(3*len(Vel))
    #find velocity scale factor
    VScale = np.sqrt(1. + (dt/tau)*((T/Tinst) - 1))
    Vel = Vel * VScale
    return Vel


def InitVelocities(N, T):
    """Returns an initial random velocity set.
Input:
    N: number of atoms
    T: target temperature
Output:
    Vel: (N,3) array of atomic velocities
"""
    Vel = np.random.rand(N, 3)
    Vel = RescaleVelocities(Vel, T)
    return Vel


def InitAccel(Pos, AtomType):
    """Returns the initial acceleration array.
Input:
    Pos: (N,3) array of atomic positions
    AtomType (N) array of atom type polymer vs. solvent
Output:
    Accel: (N,3) array of acceleration vectors
"""
    #global variables
    global M, L, Cut
    #get the acceleration from the forces
    Forces = np.zeros_like(Pos)
    PEnergy, Accel = ex3lib.calcenergyforces(Pos, M, L, Cut, Forces, AtomType)
    return Accel


def InstTemp(Vel):
    """Returns the instantaneous temperature.
Input:
    Vel: (N,3) array of atomic velocities
Output:
    Tinst: float
"""
    return np.sum(Vel * Vel) / (3. * len(Vel))


# NAtom is total number of atoms
NAtom = 240
# Rho is the atom density
Rho = 0.8
# L is box side length
L = (NAtom/Rho)**(1./3.)
# Cut is the cutoff radius
Cut = 2.5
# fractional energy tolerances for line search and conjugate gradient
EFracTolLS = 1.e-8
EFracTolCG = 1.e-10
# dx is the line search step size
dx = 0.001
# T is the equilibration temperature
T = 1.
# RescaleFreq is the velocity rescaling frequency
RescaleFreq = 1
# Dim is the number of spatial dimensions of the box/simulation
Dim = 3
# dt is the time step
dt = 0.001
# Factor is the berendsen thermostat coupling constant factor (multiplier for dt)
#Factor = 100
# tau is the Berendsen thermostat coupling constant
#tau = dt * Factor
# NStepsEquilX1 is the number of MD steps in equilibration period X

Start1 = 0
NStepsEquil1 = 25000
Stop1 = Start1 + NStepsEquil1

# EquilAvg is the number of steps at the end of the equilibration period over
# which to average the total energy for rescaling purposes
EquilAvg = 10000

Start2 = Stop1
NStepsEquil2 = 5000
Stop2 = Start2 + NStepsEquil2
Steps = NStepsEquil1 + NStepsEquil2

# NStepsProd is the number of MD steps in the production period
NStepsProd = 100000
# MD production step interval at which to log the MSD
LogInterval = 100
# number of production log entries
NumLogs = NStepsProd/LogInterval

# loop over iterations R
for R in [1]:

# loop over polymer lengths M
    for M in [16]:

# loop over polymer fractions FPoly
        for FPoly in [0.4]:

    # NumPoly is the number of polymer atoms
            NumPoly = int(NAtom * FPoly)
            
    # NumSolv is the number of solvent atoms
            NumSolv = NAtom - NumPoly

    # NPoly is the number of polymer chains
            NPoly = NumPoly/M

# define which atoms are polymer and which are solvent
            AtomType = np.zeros(NAtom, int)
            AtomType[0:NumPoly] = 1

    # initially place atoms on cubic lattice
            Pos = InitPositions(NAtom, L)

    # randomize the initial positions of the atoms such that sequential atoms
    # should (in general) not be adjacent to one another - mix things up
            np.random.shuffle(Pos)

    # perform a conjugate gradient search to find stable initial positions, PEnergy
            print "starting R="+str(R)+" M="+str(M)+" FPoly= "+str(FPoly)+" conjugate gradient"
            PEnergy, Pos = ConjugateGradient(Pos, dx, EFracTolLS, EFracTolCG, AtomType)
    
            print "conjugate gradient completed"
            print "starting equilibration period"
            
            # generate initial velocities
            Vel = InitVelocities(NAtom, T)
            
            # compute initial KEnergy from initial velocities
            KEnergy = 0.5 * np.sum(Vel * Vel)
            
            # find initial accelerations
            Accel = InitAccel(Pos, AtomType)
            
    # logs for PE, KE, and total energy during equilibration period
            PELogEq = np.zeros(Steps, float)
            KELogEq = np.zeros(Steps, float)
            ELogEq = np.zeros(Steps, float)
    
    # create an equilibration log file called "rand R# M# equil.txt"        
            f = file("solution R"+str(R)+" fpoly"+str(FPoly)+" M"+str(M)+" equil.txt", "w")
    #        f = file("berendsen 150 30k steps rescale 10k steps.txt", "w")
    
    # equilibration period, velocity rescaling ON, Berendsen time const = 200dt
            Factor = 200
            tau = dt * Factor
            for i in range(Start1, Stop1):
    #            print "M=%i Equil1 Step " %(M) + str(i)
                if (i % RescaleFreq == 0):
    #                Vel = RescaleVelocities(Vel, T)
                    Vel = Berendsen(Vel, T, dt, tau)
                    if UseVisual:
                        if atomvis.Initialized:
                            #update the positions
                            atomvis.Update(Pos)
                        else:
                            #initialize the visualization window
                            atomvis.Init(Pos)
                # do the velocity Verlet integrator to find new energies, positions, veloticies
                Pos, Vel, Accel, KEnergy, PEnergy = ex3lib.vvintegrate(Pos, Vel, Accel, M, L, Cut, dt, AtomType)
    # log the potential energy, kinetic energy, total energy during equilibration,
    # write to equilibration log file
                PELogEq[i] = PEnergy
                KELogEq[i] = KEnergy
                ELogEq[i] = PEnergy + KEnergy
                f.write("Equil Step # %06i KE= %09.3f PE= %09.3f TotalE= %09.3f\n" %(i+1, KELogEq[i], PELogEq[i], ELogEq[i]))
    
            print "equilibration w/ Berendsen thermostat completed"
    
    # perform final velocity rescaling to set instantaneous total E to the average
    # total E of the last EquilAvg steps of the equilibration period
            TargetE = ELogEq[(NStepsEquil1-EquilAvg):NStepsEquil1].mean()
            CurrentE = KEnergy + PEnergy
            DeltaE = TargetE - CurrentE
            EScalingFactor = (DeltaE+KEnergy)/KEnergy
            VScalingFactor = EScalingFactor**(1./2.)
            Vel = Vel * VScalingFactor
    
    # equilibration period, velocity rescaling OFF
            for i in range(Start2, Stop2):
                Pos, Vel, Accel, KEnergy, PEnergy = ex3lib.vvintegrate(Pos, Vel, Accel, M, L, Cut, dt, AtomType)
                PELogEq[i] = PEnergy
                KELogEq[i] = KEnergy
                ELogEq[i] = PEnergy + KEnergy
                f.write("Equil Step # %06i KE= %09.3f PE= %09.3f TotalE= %09.3f\n" %(i+1, KELogEq[i], PELogEq[i], ELogEq[i]))
    
            print "equilibration w/o thermostat completed"
    
    # close equilibration log file        
            f.close()
    
            names = ['O' for i in range(M)] + ['C' for i in range(NumPoly - M)] + ['H' for i in range(NAtom - NumPoly)]
            Pdb = atomwrite.pdbfile("anim.pdb", AtomNames = names)
            Pdb.write(Pos)
            
    # production run
            if Production:
    # Pos0 will be the positions of all atoms at the beginning of production
    # for the purposes of MSD calculation
                Pos0 = Pos.copy()
    # log files for system total, potential, and kinetic energy
    # not broken up into polymer and solvent components
                ELog = np.zeros(NumLogs, float)
                PELog = np.zeros(NumLogs, float)
                KELog = np.zeros(NumLogs, float)
    # MeanSqrDspAll is log file containing the MSD of each atom at each log point
                MeanSqrDspAll = np.zeros((NumLogs, NAtom), float)
    # MeanSqrDsp is the all-atom-average MSD at each log point
                PolyMeanSqrDsp = np.zeros(NumLogs, float)
                SolvMeanSqrDsp = np.zeros(NumLogs, float)
    
    # MD production loop
                for i in range(0, NStepsProd):
                    Pos, Vel, Accel, KEnergy, PEnergy = ex3lib.vvintegrate(Pos, Vel, Accel, M, L, Cut, dt, AtomType)
                    if (i % 200 == 0):
                        Pdb.write(Pos)
    # logging routine, happens every LogInterval MD steps
                    if (i % LogInterval == 0):
                        k = i/LogInterval
                        if UseVisual:
                            if atomvis.Initialized:
                                #update the positions
                                atomvis.Update(Pos)
                            else:
                                #initialize the visualization window
                                atomvis.Init(Pos)
    # store the energies in memory
                        ELog[k] = KEnergy + PEnergy
                        PELog[k] = PEnergy
                        KELog[k] = KEnergy
    # store MSD for each atom
                        for j in range(0, NAtom):
                            Displacement = Pos[j,:] - Pos0[j,:]
                            MeanSqrDspAll[k,j] = sum(Displacement*Displacement)
    # after production run is finished, compute all-atom average MSDs for polymer
    # and solvent components 
                PolyMeanSqrDsp = MeanSqrDspAll[:,:NumPoly].mean(1)
                SolvMeanSqrDsp = MeanSqrDspAll[:,NumPoly:].mean(1)
    
    # log file for production run
                f = file("solution R"+str(R)+" fpoly"+str(FPoly)+" M"+str(M)+" prod.txt", "w")  
    
                for i in range(0, NumLogs):
                    f.write("Step "+str(i*LogInterval))
                    f.write(" PolyMSD= %f SolvMSD = %f KE= %f PE= %f E= %f\n" %(PolyMeanSqrDsp[i], SolvMeanSqrDsp[i], KELog[i], PELog[i], ELog[i]))
    
                f.close()
    
                print "R="+str(R)+" M="+str(M)+" FPoly="+str(FPoly)+" production completed"
        
            Pdb.close()